/*
 * Copyright (C) 2003,2004,2005 Jason Woodward <woodwardj at jaos dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _GNU_SOURCE
  #define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
#include <regex.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <openssl/evp.h>
#include <libintl.h>
#include <locale.h>
#include <sys/statvfs.h>
#include <sys/mman.h>


enum action {
  USAGE = 0, UPDATE, INSTALL, REMOVE, SHOW, SEARCH, UPGRADE,
  LIST, INSTALLED, CLEAN, SHOWVERSION, AUTOCLEAN, AVAILABLE
};

#define UPDATE_OPT 'u'
#define UPGRADE_OPT 'g'
#define INSTALL_OPT 'i'
#define REMOVE_OPT 'r'
#define SHOW_OPT 's'
#define SEARCH_OPT 'e'
#define LIST_OPT 't'
#define INSTALLED_OPT 'd'
#define CLEAN_OPT 'c'
#define DOWNLOAD_ONLY_OPT 'o'
#define SIMULATE_OPT 'm'
#define VERSION_OPT 'v'
#define NO_PROMPT_OPT 'b'
#define REINSTALL_OPT 'n'
#define IGNORE_EXCLUDES_OPT 'x'
#define NO_MD5_OPT '5'
#define DIST_UPGRADE_OPT 'h'
#define HELP_OPT 'l'
#define IGNORE_DEP_OPT 'p'
#define NO_DEP_OPT 'q'
#define PRINT_URIS_OPT 'P'
#define SHOW_STATS_OPT 'S'
#define CONFIG_OPT 'C'
#define AUTOCLEAN_OPT 'a'
#define OBSOLETE_OPT 'O'
#define AVAILABLE_OPT 'A'

#define DEBUG 0
#define DO_NOT_UNLINK_BAD_FILES 1

#define _(text) gettext(text)



#define MAX_REGEX_PARTS 10
#define SLACK_BASE_SET_REGEX "^./slackware/a$"

typedef enum {
#if !defined(FALSE) && !defined(TRUE)
FALSE = 0, TRUE
#else
JFALSE = FALSE, JTRUE = TRUE
#endif
} BOOL_T;

typedef struct {
  regex_t regex;
  size_t nmatch;
  regmatch_t pmatch[MAX_REGEX_PARTS];
  int reg_return;
} sg_regex;

FILE *open_file(const char *file_name,const char *mode);
int init_regex(sg_regex *regex_t, const char *regex_string);
void execute_regex(sg_regex *regex_t,const char *string);
void free_regex(sg_regex *regex_t);
void create_dir_structure(const char *dir_name);
/* generate an md5sum of filehandle */
void gen_md5_sum_of_file(FILE *f,char *result_sum);

/* Ask the user to answer yes or no.
 * return 1 on yes, 0 on no, else -1.
 */
int ask_yes_no(const char *format, ...);
char *str_replace_chr(const char *string,const char find, const char replace);
__inline void *slapt_malloc(size_t s);
__inline void *slapt_calloc(size_t n,size_t s);


#define SOURCE_TOKEN "SOURCE="
#define WORKINGDIR_TOKEN "WORKINGDIR="
#define WORKINGDIR_TOKEN_LEN 256
#define EXCLUDE_TOKEN "EXCLUDE="

struct exclude_list {
  char **excludes;
  unsigned int count;
};

struct source_list {
  char **url;
  unsigned int count;
};

typedef struct {
  struct source_list *sources;
  char working_dir[WORKINGDIR_TOKEN_LEN];
  struct exclude_list *exclude_list;
  BOOL_T download_only;
  BOOL_T dist_upgrade;
  BOOL_T simulate;
  BOOL_T no_prompt;
  BOOL_T re_install;
  BOOL_T ignore_excludes;
  BOOL_T no_md5_check;
  BOOL_T ignore_dep;
  BOOL_T disable_dep_check;
  BOOL_T print_uris;
  BOOL_T dl_stats;
  BOOL_T remove_obsolete;
  int(*progress_cb)(void *,double,double,double,double);

} rc_config;

/*
  read the configuration from file_name
  returns (rc_config *) or NULL
*/
rc_config *read_rc_config(const char *file_name);

/*
  check that working_dir exists or make it if permissions allow
*/
void working_dir_init(const rc_config *global_config);

/*
  free rc_config structure
*/
void free_rc_config(rc_config *global_config);

/*
  add an exclude expression to the exclude list.
  commonly called with global_config->exclude_list
*/
void add_exclude(struct exclude_list *list,const char *e);

/*
  add an source expression to the source list.
  commonly called with global_config->source_list
*/
void add_source(struct source_list *list,const char *s);


#define PKG_PARSE_REGEX "(.*{1,})\\-(.*[\\.\\-].*[\\.\\-].*).tgz[ ]{0,}$"
#define PKG_NAMEVER "(.*{1,})\\-(.*[\\.\\-].*[\\.\\-].*)"
#define PKG_VER "(.*)[\\.\\-](.*)[\\.\\-](.*)"
#define PKG_NAME_PATTERN "^PACKAGE NAME:[ ]{1,}(.*{1,})\\-(.*[\\.\\-].*[\\.\\-].*).tgz[ ]{0,}$"
#define PKG_MIRROR_PATTERN "^PACKAGE MIRROR:[ ]+(.*)$"
#define PKG_LOCATION_PATTERN "^PACKAGE LOCATION:[ ]+(.*)$"
#define PKG_SIZEC_PATTERN "^PACKAGE SIZE [(]+compressed[)]{1,}:[ ]{1,}([0-9]{1,}) K$"
#define PKG_SIZEU_PATTERN "^PACKAGE SIZE [(]+uncompressed[)]{1,}:[ ]{1,}([0-9]{1,}) K$"
#define PKG_LOG_SIZEC_PATTERN "^COMPRESSED PACKAGE SIZE:[ ]{1,}([0-9]{1,}) K$"
#define PKG_LOG_SIZEU_PATTERN "^UNCOMPRESSED PACKAGE SIZE:[ ]{1,}([0-9]{1,}) K$"
#define PKG_LOG_DIR "/var/log/packages"
#define ROOT_ENV_NAME "ROOT"
#define ROOT_ENV_LEN 255
#define PKG_LOG_PATTERN "^(.*{1,})\\-(.*[\\.\\-].*[\\.\\-].*)"
#define MD5SUM_REGEX "([a-zA-Z0-9]{1,})[ ]{1,}([a-zA-Z0-9\\/._\\-]{1,})\\/(.*{1,})\\-(.*[\\.\\-].*[\\.\\-].*).tgz$"
#define REQUIRED_REGEX "^[ ]{0,}([a-zA-Z0-9\\+_\\-]+)[ ]{0,}([\\<\\=\\>]+){0,}[ ]{0,}([a-zA-Z0-9\\.\\_\\-]+){0,}[ ]{0,}$"
#define MD5_STR_LEN 34
#define MD5_CHECKSUM_FAILED -100
#define PKG_LIST "PACKAGES.TXT"
#define PKG_LIST_L "package_data"
#define PATCHES_LIST "patches/PACKAGES.TXT"
#define PATCHDIR "patches/"
#define REMOVE_CMD "/sbin/removepkg "
#define INSTALL_CMD "/sbin/installpkg "
#define UPGRADE_CMD "/sbin/upgradepkg --reinstall "
#define CHECKSUM_FILE "CHECKSUMS.md5"
#define HEAD_FILE_EXT ".head"
#define MAX_MMAP_SIZE 1024

typedef struct {
  char *name;
  char *version;
  char *mirror;
  char *location;
  unsigned int size_c;
  unsigned int size_u;
  char *description;
  char *required;
  char *conflicts;
  char *suggests;
  char md5[MD5_STR_LEN];
} pkg_info_t;

struct pkg_list {
  pkg_info_t **pkgs;
  unsigned int pkg_count;
  BOOL_T free_pkgs;
};

typedef struct {
  pkg_info_t *installed;
  pkg_info_t *upgrade;
} pkg_upgrade_t;

struct pkg_upgrade_list {
  pkg_upgrade_t **pkgs;
  unsigned int pkg_count;
};

struct pkg_version_parts {
  char **parts;
  unsigned int count;
};

typedef struct {
  char *pkg;
  char *error;
} pkg_err_t;

struct pkg_err_list {
  unsigned int err_count;
  pkg_err_t **errs;
};

/* returns an empty package structure */
__inline pkg_info_t *init_pkg(void);
/* frees the package structure */
void free_pkg(pkg_info_t *pkg);

/* create an empty package list */
struct pkg_list *init_pkg_list(void);
/* add a package to a package list */
void add_pkg_to_pkg_list(struct pkg_list *list,pkg_info_t *pkg);
/* free package list */
void free_pkg_list(struct pkg_list *);

/* update the local package cache */
int update_pkg_cache(const rc_config *global_config);
/* write pkg data to disk */
void write_pkg_data(const char *source_url,FILE *d_file,struct pkg_list *pkgs);
/* parse the PACKAGES.TXT file */
struct pkg_list *parse_packages_txt(FILE *);
/* return a list of available packages */
struct pkg_list *get_available_pkgs(void);
/* retrieve list of installed pkgs */
struct pkg_list *get_installed_pkgs(void);

/* retrieve the newest package from package list */
pkg_info_t *get_newest_pkg(struct pkg_list *,const char *);
/* get the exact package */
pkg_info_t *get_exact_pkg(struct pkg_list *list,const char *name,
                          const char *version);
/* lookup package by details */
pkg_info_t *get_pkg_by_details(struct pkg_list *list,char *name,
                               char *version,char *location);
/* search package list with pattern */
struct pkg_list *search_pkg_list(struct pkg_list *available,
                                 const char *pattern);


/*
  install package by calling installpkg
  returns 0 on success, -1 on error
*/
int install_pkg(const rc_config *,pkg_info_t *);
/*
  upgrade package by calling upgradepkg
  returns 0 on success, -1 on error
*/
int upgrade_pkg(const rc_config *global_config,pkg_info_t *installed_pkg,
                pkg_info_t *pkg);
/*
  remove package by calling removepkg
  returns 0 on success, -1 on error
*/
int remove_pkg(const rc_config *,pkg_info_t *);


/*
  generate a short description, returns (char *) on success or NULLon error
  caller responsible for freeing the returned data
*/
char *gen_short_pkg_description(pkg_info_t *);
/*
  generate the filename from the url
  caller responsible for freeing the returned data
*/
char *gen_filename_from_url(const char *url,const char *file);
/*
  generate the package file name
  caller responsible for freeing the returned data
*/
char *gen_pkg_file_name(const rc_config *global_config,pkg_info_t *pkg);
/*
  generate the head cache filename
  caller responsible for freeing the returned data
*/
char *gen_head_cache_filename(const char *filename_from_url);
/*
  generate the download url for a package
  caller responsible for freeing the returned data
*/
char *gen_pkg_url(pkg_info_t *pkg);
/*
  exclude pkg based on pkg name
  returns 1 if package is present in the exclude list, 0 if not present
*/
int is_excluded(const rc_config *,pkg_info_t *);
/*
  package is already downloaded and cached, md5sum if applicable is ok
  returns 0 if download is cached, -1 if not
*/
int verify_downloaded_pkg(const rc_config *global_config,pkg_info_t *pkg);
/*
  fill in the md5sum of the package
*/
void get_md5sum(pkg_info_t *pkg,FILE *checksum_file);
/*
  find out the pkg file size (post download)
*/
size_t get_pkg_file_size(const rc_config *global_config,pkg_info_t *pkg);

/*
  compare package versions
  returns just like strcmp,
    > 0 if a is greater than b
    < 0 if a is less than b
    0 if a and b are equal
*/
int cmp_pkg_versions(char *a, char *b);

/*
  resolve dependencies
  returns 0 on success, -1 on error setting conflict_err and missing_err
  (usually called with transaction->conflict_err and transaction->missing_err)
*/
int get_pkg_dependencies(const rc_config *global_config,
                         struct pkg_list *avail_pkgs,
                         struct pkg_list *installed_pkgs,pkg_info_t *pkg,
                         struct pkg_list *deps,
                         struct pkg_err_list *conflict_err,
                         struct pkg_err_list *missing_err);
/*
  return list of package conflicts
*/
struct pkg_list *get_pkg_conflicts(struct pkg_list *avail_pkgs,
                                   struct pkg_list *installed_pkgs,
                                   pkg_info_t *pkg);
/*
  return list of packages required by
*/
struct pkg_list *is_required_by(const rc_config *global_config,
                                struct pkg_list *avail, pkg_info_t *pkg);

/*
  empty packages from cache dir
*/
void clean_pkg_dir(const char *dir_name);
/*
  clean out old outdated packages in the cache that are no longer available
  in the current source lists (ie are not downloadable)
*/
void purge_old_cached_pkgs(const rc_config *global_config,char *dir_name,
                           struct pkg_list *avail_pkgs);

/*
  make a copy of a package (needs to be freed with free_pkg)
*/
pkg_info_t *copy_pkg(pkg_info_t *dst,pkg_info_t *src);

/*
  package error handling api to handle errors within core functions
*/
struct pkg_err_list *init_pkg_err_list(void);
void add_pkg_err_to_list(struct pkg_err_list *l,
                         const char *pkg,const char *err);
int search_pkg_err_list(struct pkg_err_list *l,
                        const char *pkg, const char *err);
void free_pkg_err_list(struct pkg_err_list *l);


/*
  this is the main download routine, returns 0 on success.  On error returns
  -1 or CURLE_HTTP_RANGE_ERROR
*/
int download_data(FILE *fh,const char *url,size_t bytes,
                  const rc_config *global_config);

/*
  retrieves the head data for the url, returns (char *) or NULL on error
*/
char *head_request(const char *url);

/*
  this fills FILE with data from url, used for PACKAGES.TXT and CHECKSUMS
  returns 0 on success, on error -1 or CURLE_HTTP_RANGE_ERROR
*/
int get_mirror_data_from_source(FILE *fh,const rc_config *global_config,
                                const char *base_url,const char *filename);

/*
  download pkg, calls download_data
  returns 0 on success, on error -1 or CURLE_HTTP_RANGE_ERROR
*/
int download_pkg(const rc_config *global_config,pkg_info_t *pkg);

/* callback for curl progress */
char spinner(void);

/*
  this is the default progress callback if global_config->progress_cb == NULL
*/
int progress_callback(void *clientp, double dltotal, double dlnow,
                      double ultotal, double ulnow);

/*
  do a head request on the mirror data to find out if it's new
  returns (char *) or NULL
*/
char *head_mirror_data(const char *wurl,const char *file);
/*
  clear head cache storage file
*/
void clear_head_cache(const char *cache_filename);
/*
  cache the head request
*/
void write_head_cache(const char *cache, const char *cache_filename);
/*
  read the cached head request
  returns (char *) or NULL
*/
char *read_head_cache(const char *cache_filename);


/*
  this defines the max length of the transaction summary lines, hopefully
  someday this will be replaced with a more dynamic solution.
*/
#define MAX_LINE_LEN 80

struct suggests {
  char **pkgs;
  unsigned int count;
};

typedef struct {
  union { pkg_info_t *i; pkg_upgrade_t *u; } pkg;
  unsigned int type; /* this is enum action defined in main.h */
} queue_i;

typedef struct {
  queue_i **pkgs;
  unsigned int count;
} queue_t;

typedef struct {
  struct pkg_list *install_pkgs;
  struct pkg_upgrade_list *upgrade_pkgs;
  struct pkg_list *remove_pkgs;
  struct pkg_list *exclude_pkgs;
  struct suggests *suggests;
  struct pkg_err_list *conflict_err;
  struct pkg_err_list *missing_err;
  queue_t *queue;
} transaction_t;

/* fill in transaction structure with defaults */
void init_transaction(transaction_t *);
/*
  download and install/remove/upgrade packages as defined in the transaction
  returns 0 on success
*/
int handle_transaction(const rc_config *,transaction_t *);

/* add package for installation to transaction */
void add_install_to_transaction(transaction_t *,pkg_info_t *pkg);
/* add package for removal to transaction */
void add_remove_to_transaction(transaction_t *,pkg_info_t *pkg);
/* add package to upgrade to transaction */
void add_upgrade_to_transaction(transaction_t *,pkg_info_t *installed_pkg,
                                pkg_info_t *upgrade_pkg);
/* add package to exclude to transaction */
void add_exclude_to_transaction(transaction_t *,pkg_info_t *pkg);
/* remove package from transaction, returns modified transaction */
transaction_t *remove_from_transaction(transaction_t *tran,pkg_info_t *pkg);

/* search transaction by package name.  returns 1 if found, 0 otherwise */
int search_transaction(transaction_t *,char *pkg_name);
/*
  search transaction by package attributes
  returns 1 if found, 0 otherwise
*/
int search_transaction_by_pkg(transaction_t *tran,pkg_info_t *pkg);
/*
  searches the upgrade list of the transaction for the present of the package
  returns 1 if found, 0 if not found
*/
int search_upgrade_transaction(transaction_t *tran,pkg_info_t *pkg);

/*
  add dependencies for package to transaction, returns -1 on error, 0 otherwise
*/
int add_deps_to_trans(const rc_config *global_config, transaction_t *tran,
                      struct pkg_list *avail_pkgs,
                      struct pkg_list *installed_pkgs, pkg_info_t *pkg);

/*
  check to see if a package has a conflict already present in the transaction
  returns conflicted package or NULL if none
*/
pkg_info_t *is_conflicted(transaction_t *tran, struct pkg_list *avail_pkgs,
                          struct pkg_list *installed_pkgs, pkg_info_t *pkg);

/*
  generate a list of suggestions based on the current packages in the transaction
*/
void generate_suggestions(transaction_t *tran);

/* free the transaction structure and it's members */
void free_transaction(transaction_t *);
